from datetime import timedelta

from django.template.defaultfilters import floatformat
from django.utils.html import format_html
from django.utils.safestring import SafeString

import django_tables2 as tables

from .models import EventAttendee, EventExt, Member


class DurationColumn(tables.Column):
    def render(self, value: timedelta):
        if value is None:
            return None
        return floatformat(value.total_seconds() / 60 / 60, -2)

    def value(self, value: timedelta):
        if value is None:
            return None
        return value.total_seconds() / 60 / 60


class EventTable(tables.Table):
    title = tables.TemplateColumn(
        template_code=(
            '<a title="MembershipWorks" href="https://membershipworks.com/admin/#!event/admin/{{ record.url }}">{{ value }}</a> '
            '<a title="Admin" href="{% url "admin:membershipworks_eventext_change" record.pk %}"><i class="bi bi-pencil-square"></i></a> '
            '<a title="Details" href="{% url "membershipworks:event-detail" record.pk %}"><i class="bi bi-receipt"></i></a> '
        ),
        accessor="unescaped_title",
    )
    occurred = tables.BooleanColumn(visible=False)
    start = tables.DateColumn("N d, Y")
    duration = DurationColumn()
    person_hours = DurationColumn()
    meetings = tables.Column()
    gross_revenue = tables.Column()
    total_due_to_instructor = tables.Column()
    net_revenue = tables.Column()
    invoice__date_submitted = tables.DateColumn(verbose_name="Invoice Submitted")
    invoice__date_paid = tables.DateColumn(verbose_name="Invoice Paid")

    class Meta:
        model = EventExt
        fields = (
            "title",
            "occurred",
            "start",
            "instructor",
            "category",
            "count",
            "cap",
            "meetings",
            "duration",
            "person_hours",
            "gross_revenue",
            "total_due_to_instructor",
            "net_revenue",
        )
        row_attrs = {
            "class": lambda record: (
                "" if record.occurred else "text-decoration-line-through table-danger"
            )
        }


class EventRegistrationsTable(tables.Table):
    ticket_count = tables.Column(empty_values=())
    name = tables.Column(accessor="Full name")
    email = tables.EmailColumn(accessor="Email")
    phone = tables.Column(accessor="Phone")
    emergency_contact_name = tables.Column(accessor="Emergency Contact Name:")
    emergency_contact_phone_number = tables.Column(
        accessor="Emergency Contact Phone Number:"
    )
    emergency_contact_relation = tables.Column(accessor="Emergency Contact Relation:")

    def render_ticket_count(self, record):
        return sum(int(v) for k, v in record.items() if k.startswith("Ticket: "))

    class Meta:
        row_attrs = {
            "class": lambda table, record: (
                ""
                if table.render_ticket_count(record) > 0
                else "text-decoration-line-through table-danger"
            )
        }


class EventSummaryTable(tables.Table):
    event_count = tables.Column("Events")
    canceled_event_count = tables.Column("Canceled Events")
    count__sum = tables.Column("Tickets")
    instructor__count = tables.Column("Unique Instructors")
    meetings__sum = tables.Column("Meetings")
    duration__sum = DurationColumn("Class Hours")
    person_hours__sum = DurationColumn("Person Hours")
    gross_revenue__sum = tables.Column("Gross Revenue")
    total_due_to_instructor__sum = tables.Column("Total Due to Instructor")
    net_revenue__sum = tables.Column("Net Revenue")


class UserEventTable(EventTable):
    title = tables.Column(linkify=True, accessor="unescaped_title")
    instructor = None
    person_hours = None
    gross_revenue = None
    net_revenue = None


class InvoiceMoneyColumn(tables.columns.Column):
    def render(self, value):
        return f"${super().render(value):.2f}"


class InvoiceMoneyFooterColumn(InvoiceMoneyColumn):
    def render_footer(self, bound_column, table):
        value = getattr(table.event, bound_column.accessor)
        if value is not None:
            return f"${value:.2f}"
        else:
            return bound_column.default


class InvoiceTable(tables.Table):
    def __init__(self, *args, **kwargs):
        self.event = kwargs.pop("event")
        super().__init__(*args, **kwargs)

    @staticmethod
    def _math_header(name: str, formula: str) -> SafeString:
        return format_html(
            '{} <div class="text-nowrap font-monospace fw-light">[{}]</div>',
            name,
            formula,
        )

    label = tables.Column("Ticket Type", footer="Subtotals")
    list_price = InvoiceMoneyColumn("Ticket Price")
    actual_price = InvoiceMoneyColumn(_math_header("Actual Price", "P"))
    quantity = tables.Column(
        _math_header("Quantity", "Q"),
        footer=lambda table: table.event.quantity,
    )
    amount = InvoiceMoneyFooterColumn(_math_header("Amount", "A=P*Q"))
    materials = InvoiceMoneyFooterColumn(
        _math_header("CMS Collected Materials Fee", "M=m*Q")
    )
    amount_without_materials = InvoiceMoneyFooterColumn(
        _math_header("Event Revenue Base", "B=A-M")
    )
    instructor_revenue = InvoiceMoneyFooterColumn(
        _math_header("Instructor Percentage Revenue", "R=B*I")
    )
    instructor_amount = InvoiceMoneyFooterColumn(
        _math_header("Amount Due to Instructor", "R+M")
    )

    class Meta:
        attrs = {
            "class": "table table-sm mx-auto w-auto",
            "tbody": {"class": "table-group-divider"},
            "tfoot": {"class": "table-group-divider"},
        }
        orderable = False


class EventAttendeeTable(tables.Table):
    class Meta:
        model = EventAttendee
        fields = ("name", "email")


class MissingPaperworkTable(tables.Table):
    policy_agreement = tables.BooleanColumn()
    authorize_charge = tables.BooleanColumn()

    class Meta:
        model = Member
        fields = [
            "first_name",
            "last_name",
            "membership",
            "billing_method",
            "join_date",
            "membership_agreement_signed_and_on_file_date",
            "waiver_form_signed_and_on_file_date",
            "policy_agreement",
            "authorize_charge",
        ]
